"""
The LighBlueService should be used as API wrapper. It makes http query based
on user's requirements.
"""

import json
import logging

from lightblue.common import retry_session

LOGGER = logging.getLogger('lightblue')


class LightBlueService(object):
    """"
        Class for interacting with lightBlue API
    """

    def __init__(
        self,
        data_url,
        metadata_url,
        ssl_certificate=None,
        ssl_verify=True,
        custom_session=None,
    ):
        self.data_url = data_url.rstrip('/')
        self.metadata_url = metadata_url.rstrip('/')
        self.ssl_certificate = ssl_certificate
        if custom_session is None:
            self.session = retry_session()
            self.session.verify = ssl_verify
            if self.ssl_certificate is not None:
                self.session.cert = self.ssl_certificate
        else:
            self.session = custom_session

    @staticmethod
    def log_response(response):
        """
        Logging API calls response
        Args:
            response: API call response
        """
        log_response = {'statusCode': response.status_code}
        try:
            response_data = response.json()
            log_response['status'] = response_data.get('status')
            log_response['matchCount'] = response_data.get('matchCount')
            log_response['modifiedCount'] = response_data.get('modifiedCount')
            # data should be excluded from log message - include only
            # data errors
            if response_data.get('status') != 'COMPLETE':
                log_response['dataErrors'] = response_data.get('dataErrors')
                log_response['errors'] = response_data.get('errors')
        except ValueError:
            log_response['text_response'] = response.text

        LOGGER.debug("LightBlue response - %s", log_response)
        return log_response

    def get_schema(self, entity_name, version):
        """
        Get schema for specific entity
        Args:
            entity_name (str): entity name
            version (str): entity version

        Returns:
            - dict - schema of given entity

        """
        url = '{metadata_url}/{entity_name}/{version}'.format(
            metadata_url=self.metadata_url,
            entity_name=entity_name,
            version=version
        )
        LOGGER.debug("%s - %s", 'GET', url)
        response = self.session.get(url)
        response.raise_for_status()
        return response.json()

    def insert_data(self, entity_name, version, data):
        """
        Insert request
        Args:
            entity_name (str): entity name
            version (str/None): entity version
            data (dict/list): new lightblue documents

        Returns:
            - dict - lightblue response

        """
        url = '{data_url}/insert/{entity_name}'
        if version is not None:
            url = url + '/{version}'
        url = url.format(
            data_url=self.data_url, entity_name=entity_name, version=version
        )
        LOGGER.debug("%s - %s", 'PUT', url)
        response = self.session.put(url, json=data)
        self.log_response(response)
        if response.status_code != 200:
            LOGGER.error('Insert data failed - %s', json.dumps(data))
            return None
        return response.json()

    def delete_data(self, entity_name, version, data):
        """
        Delete data according to data query
        Args:
            entity_name (str): entity name
            version (str/None): entity version
            data (dict): data contains query

        Returns:
            - dict - lightblue response

        """

        url = '{data_url}/delete/{entity_name}'
        if version is not None:
            url = url + '/{version}'
        url = url.format(
            data_url=self.data_url, entity_name=entity_name, version=version
        )
        LOGGER.debug("%s - %s", 'POST', url)
        response = self.session.post(url, json=data)
        self.log_response(response)
        if response.status_code != 200:
            LOGGER.error('Delete data failed - %s', json.dumps(data))
            return None
        return response.json()

    def update_data(self, entity_name, version, data):
        """
        Update data according to data query and update field
        Args:
            entity_name (str): entity_name
            version (str/None): entity version
            data (dict): data contains query and update field

        Returns:
            - dict - lightblue response

        """
        url = '{data_url}/update/{entity_name}'
        if version is not None:
            url = url + '/{version}'
        url = url.format(
            data_url=self.data_url, entity_name=entity_name, version=version
        )
        LOGGER.debug("%s - %s", 'POST', url)
        response = self.session.post(url, json=data)
        self.log_response(response)
        if response.status_code != 200:
            LOGGER.error('Update data failed - %s', json.dumps(data))
            return None
        return response.json()

    def find_data(self, entity_name, version, data):
        """
        Find data according to data query
        Args:
            entity_name (str): entity name
            version (str/None): entity version
            data (dict): data contains query and projection field

        Returns:
            - dict - result of search and projection query

        """

        url = '{data_url}/find/{entity_name}'
        if version is not None:
            url = url + '/{version}'
        url = url.format(
            data_url=self.data_url, entity_name=entity_name, version=version
        )
        LOGGER.debug("%s - %s", 'POST', url)
        response = self.session.post(url, json=data)
        self.log_response(response)
        if response.status_code != 200:
            LOGGER.error('Find data failed - %s', json.dumps(data))
            return None
        return response.json()



from unittest import TestCase

from lightblue.entity import LightBlueEntity
from . import FakeLightblueService

try:
    from unittest.mock import Mock, patch, call
except ImportError:
    from mock import Mock, patch, call


class TestLightBlueEntity(TestCase):
    """
    Test cases for LightBlueEntity class
    """
    test_docstring_prefix = "Generic entity - "

    def shortDescription(self):  # noqa
        """Override nosetest docstrings."""
        doc = self.test_docstring_prefix + self._testMethodDoc
        return doc or None

    def setUp(self):
        self.fake_lightblue_service = FakeLightblueService()
        self.lb_entity = LightBlueEntity(self.fake_lightblue_service,
                                         'fake-name', 'fake_version')

    def test_check_response_success(self):
        """
        Test response message is complete
        """
        response = {
            'status': 'COMPLETE'
        }
        result = self.lb_entity.check_response(response)
        self.assertTrue(result)

    def test_check_response_failure(self):
        """
        Test response message is missing some values
        """
        result = self.lb_entity.check_response(None)
        self.assertFalse(result)

        # no status in response
        response = {
            'fake-key': 'value'
        }
        result = self.lb_entity.check_response(response)
        self.assertFalse(result)

        # status is not equal to COMPLETE
        response = {
            'status': 'PARTIAL'
        }
        result = self.lb_entity.check_response(response)
        self.assertFalse(result)

    def test_get_schema_no_version(self):
        """
        Test get entity schema - no version provided
        """
        self.lb_entity.version = None
        with self.assertRaises(ValueError):
            self.lb_entity.get_schema()

    def test_get_schema_class_version(self):
        """
        Test get entity schema - class version used
        """
        self.lb_entity.get_schema()
        self.fake_lightblue_service.get_schema.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version
        )

    def test_get_schema_given_version(self):
        """
        Test get entity schema - arg version used
        """
        self.lb_entity.get_schema('my_version')
        self.fake_lightblue_service.get_schema.assert_called_once_with(
            self.lb_entity.entity_name, 'my_version'
        )

    def test_insert_data(self):
        """
        Test insert new data
        """
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'data': 'fake_data',
            'version': self.lb_entity.version,
            'projection': {
                'field': '*',
                'include': True,
                'recursive': True
            }
        }
        self.lb_entity.insert_data('fake_data')
        self.fake_lightblue_service.insert_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    def test_delete_all(self):
        """
        Test delete all documents
        """
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'version': self.lb_entity.version,
            'query': {
                'field': 'objectType',
                'op': '=',
                'rvalue': self.lb_entity.entity_name
            }
        }
        self.lb_entity.delete_all()
        self.fake_lightblue_service.delete_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    def test_delete_item(self):
        """
        Test delete document based on query
        """
        query = 'fake_query'
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'version': self.lb_entity.version,
            'query': query
        }
        self.lb_entity.delete_item(query)
        self.fake_lightblue_service.delete_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    def test_update_item(self):
        """
        Test update document based on query and given data
        """
        query = 'fake_query'
        data = 'fake_data'
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'version': self.lb_entity.version,
            'query': query,
            'update': data
        }
        self.lb_entity.update_item(query, data)
        self.fake_lightblue_service.update_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    def test_find_item(self):
        """
        Test find document based on query
        """
        query = 'fake_query'
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'version': self.lb_entity.version,
            'query': query,
            'projection': {
                'field': '*',
                'include': True,
                'recursive': True
            },
            'from': 10,
            'maxResults': 10,
        }
        self.lb_entity.find_item(query, from_=10, max_results=10)
        self.fake_lightblue_service.find_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    def test_find_all(self):
        """
        Test find all documents
        """
        query = {
            'field': 'objectType',
            'op': '=',
            'rvalue': self.lb_entity.entity_name
        }
        expected_data = {
            'objectType': self.lb_entity.entity_name,
            'version': self.lb_entity.version,
            'query': query,
            'projection': {
                'field': '*',
                'include': True,
                'recursive': True
            },
            'from': 10,
            'maxResults': 10,
        }
        self.lb_entity.find_all(from_=10, max_results=10)
        self.fake_lightblue_service.find_data.assert_called_once_with(
            self.lb_entity.entity_name, self.lb_entity.version, expected_data
        )

    @patch('lightblue.entity.LightBlueEntity.check_response')
    def test_find_paginated(self, mock_check_response):
        find_func = Mock()
        find_func.side_effect = [
            {
                'processed': ['value 1', 'value 2']
            },
            {
                'processed': ['value 3']
            },
            {
                'processed': []
            }
        ]
        mock_check_response.return_value = True
        result = self.lb_entity.find_paginated(
                200, find_func, 'a', 'b', 'c', d='e'
        )
        self.assertEqual(result, ['value 1', 'value 2', 'value 3'])
        self.assertEqual(
            find_func.call_args_list,
            [
                call('a', 'b', 'c', d='e', from_=0, max_results=200),
                call('a', 'b', 'c', d='e', from_=200, max_results=200),
                call('a', 'b', 'c', d='e', from_=400, max_results=200),
            ]
        )

    @patch('lightblue.entity.LightBlueEntity.check_response')
    def test_find_paginated_one_page(self, mock_check_response):
        find_func = Mock()
        find_func.side_effect = [
            {
                'processed': ['value 1', 'value 2']
            },
            {
                'processed': []
            }
        ]
        mock_check_response.return_value = True
        result = self.lb_entity.find_paginated(
                100, find_func, 'a', 'b', 'c', d='e'
        )
        self.assertEqual(result, ['value 1', 'value 2'])
        self.assertEqual(find_func.call_count, 2)

    @patch('lightblue.entity.LightBlueEntity.check_response')
    def test_find_paginated_no_results(self, mock_check_response):
        find_func = Mock()
        find_func.side_effect = [
            {
                'processed': []
            }
        ]
        mock_check_response.return_value = True
        result = self.lb_entity.find_paginated(
                100, find_func, 'a', 'b', 'c', d='e'
        )
        self.assertEqual(result, [])
        self.assertEqual(find_func.call_count, 1)

    @patch('lightblue.entity.LightBlueEntity.check_response')
    def test_find_paginated_invalid_response(self, mock_check_response):
        find_func = Mock()
        find_func.side_effect = [
            {
                'processed': ['value 1', 'value 2']
            },
            {
                'invalid': 'response'
            },
            {
                'processed': ['value 3']
            }
        ]
        mock_check_response.side_effect = [True, False, True]
        result = self.lb_entity.find_paginated(
                100, find_func, 'a', 'b', 'c', d='e'
        )
        self.assertEqual(result, None)
        self.assertEqual(find_func.call_count, 2)



import logging

LOGGER = logging.getLogger('lightblue')


class LightBlueEntity(object):
    """
    Lightblue generic entity
     - all common action for all entities are specified here
     - for specific entity create new class with this class as parent
    """
    def __init__(self,
                 lightblue_service,
                 entity_name,
                 version=None,
                 *args,
                 **kwargs):

        super(LightBlueEntity, self).__init__(*args, **kwargs)
        self.service = lightblue_service
        self.entity_name = entity_name
        self.version = version

    @staticmethod
    def check_response(response):
        """
        Check LightBlue response. It always has to contain
            {
                'status': 'COMPLETE'
                ..
            }
        Args:
            response (dict): Lightblue API response

        Returns:
             bool - True if response is with no errors / False

        """
        if response is not None and \
           'status' in response and \
           response['status'] == 'COMPLETE':
            return True
        LOGGER.error('Received invalid response from Lightblue: %s', response)
        return False

    def get_schema(self, version=None):
        """
        Get schema for generic entity
        Args:
            version (str): entity version (if missing, default is used)

        Returns:
            - dict - json schema for entity

        """
        if version is None:
            if self.version is None:
                raise ValueError("No version was provided")
            return self.service.get_schema(self.entity_name, self.version)
        else:
            return self.service.get_schema(self.entity_name, version)

    def insert_data(self, data):
        """
        Insert data to generic entity (default projection is used)
        Args:
            data (dict/list): json objects for entity

        Returns:
            - dict - lightblue response

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'data': data,
            'projection': {
                'field': '*',
                'include': True,
                'recursive': True

            }
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        return self.service.insert_data(self.entity_name, self.version,
                                        lightblue_data)

    def delete_all(self):
        """
        Delete all data for generic entity
        Returns:
            - dict - lightblue response

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'query': {
                'field': 'objectType',
                'op': '=',
                'rvalue': self.entity_name
            }
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        return self.service.delete_data(self.entity_name, self.version,
                                        lightblue_data)

    def delete_item(self, query):
        """
        Delete specific object according to query
        Args:
            query (dict): query for delete

        Returns:
            - dict - lightblue response

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'query': query
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        return self.service.delete_data(self.entity_name, self.version,
                                        lightblue_data)

    def update_item(self, query, update):
        """
        Update specific object according to query and update field
        Args:
            query (dict): query for searching data
            update (dict): update field (new values for given fields)

        Returns:
            - dict - lightblue response

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'query': query,
            'update': update
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        return self.service.update_data(self.entity_name, self.version,
                                        lightblue_data)

    def find_item(self, query, projection=None, from_=None, max_results=None):
        """
        Find specific object according to query and projection field
        Args:
            query (dict): search query
            projection (list): specify field which will be returned
                               (default - return all)
            from_: from item in query
            max_results: limit results count in response

        Returns:
            - dict - result of search query

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'query': query
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        if projection is None:
            projection = {
                'field': '*',
                'include': True,
                'recursive': True
            }
        if from_ is not None:
            lightblue_data['from'] = from_
        if max_results is not None:
            lightblue_data['maxResults'] = max_results
        lightblue_data['projection'] = projection
        return self.service.find_data(self.entity_name, self.version,
                                      lightblue_data)

    def find_all(self, projection=None, from_=None, max_results=None):
        """
        Find all objects of given entity
        Args:
            projection (list): custom projection (default return all items)
            from_: from item in query
            max_results: limit results count in response

        Returns:
            - dict - result of search query

        """
        lightblue_data = {
            'objectType': self.entity_name,
            'query': {
                'field': 'objectType',
                'op': '=',
                'rvalue': self.entity_name
            }
        }
        if self.version is not None:
            lightblue_data['version'] = self.version
        if projection is None:
            projection = {
                'field': '*',
                'include': True,
                'recursive': True
            }
        if from_ is not None:
            lightblue_data['from'] = from_
        if max_results is not None:
            lightblue_data['maxResults'] = max_results
        lightblue_data['projection'] = projection
        return self.service.find_data(self.entity_name, self.version,
                                      lightblue_data)

    def find_paginated(self, page_size, find, *args, **kwargs):
        """
        Get joined 'processed' key from paginated find calls

        Args:
            page_size (int): max results per LightBlue call
            find (Callable): find function (find_item / find_all)

        Returns:
            - list of processed items from multiple find calls

        """
        kwargs.update({
            "from_": 0,
            "max_results": page_size
        })

        processed = []
        while True:
            response = find(*args, **kwargs)
            if not self.check_response(response):
                return None
            elif len(response['processed']) == 0:
                return processed

            processed.extend(response['processed'])
            kwargs['from_'] += page_size



try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock

from lightblue.service import LightBlueService


class FakeLightblueService(LightBlueService):
    def __init__(self, *args, **kwargs):
        self.data_api = Mock()
        self.metadata_api = Mock()
        self.insert_data = Mock()
        self.delete_data = Mock()
        self.update_data = Mock()
        self.find_data = Mock()
        self.get_schema = Mock()



"""LightBlueGenericSelection implementation."""

import dpath.util

from lightblue.query import LightBlueQuery


class LightBlueGenericSelection(LightBlueQuery):
    """
    Representation of an item in a collection.

    (isn't specific to any LightBlue entity)

    Attributes:
        interface (lightblue.entity.LightBlueEntity):
            wrapper to query a LightBlue method
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a LightBlueGenericSelection object.

        Args:
            interface (lightblue.entity.LightBlueEntity):
                reference to LightBlueEntity object
            *args: Description
            **kwargs: Description
        """
        self.interface = kwargs.pop('interface')

        super(LightBlueGenericSelection, self).__init__(
            self.interface, *args, **kwargs)

    def _postprocessing(self,
                        result,
                        check_response=True,
                        selector=None,
                        count=None,
                        fallback=None,
                        postprocess=None):
        """
        Post-process the response from LightBlue.

        Args:
            result (dict): response from LightBlue
            check_response (bool, optional):
                return fallback if .check_response() failed
            selector (None, optional):
                dpath selector to get the result of query
            count (None, optional):
                - matchCount have to equal count if count is int
                - matchCount < count[0] and  matchCount < count[2]
                  in case of tuple
            fallback (None, optional): value to return if any check fails
            postprocess (None, optional):
                the result will be passed as an argument to the function

        Returns:
            object:
                - whole response
                - selected part of the response
                - fallback provided to the post-process
        """
        # check response
        if check_response:
            if not self.interface.check_response(result):
                return fallback
        # check count
        if count:
            if isinstance(count, tuple):
                if result['matchCount'] < count[0]:
                    return fallback
                if len(count) > 1:
                    if result['matchCount'] > count[1]:
                        return fallback
            elif isinstance(count, int):
                if result['matchCount'] != count:
                    return fallback
        # return value (by selector) or result
        if selector:
            try:
                result = dpath.util.get(result, selector)
            except ValueError:
                result = dpath.util.values(result, selector)
        if postprocess:
            postprocess_result = postprocess(result)
            if postprocess_result is None:
                return fallback
            else:
                return postprocess_result
        else:
            return result

    def find(self, *args, **kwargs):
        """
        Postprocessing wrapper over find method.

        Args:
            *args: arguments to pass to _postprocessing()
            **kwargs: arguments to pass to _postprocessing()

        Returns:
            object: as returned by _postprocessing()
        """
        result = super(LightBlueGenericSelection, self).find()
        return self._postprocessing(result, *args, **kwargs)

    def update(self, *args, **kwargs):
        """
        Postprocessing wrapper over update method.

        Args:
            *args: arguments to pass to _postprocessing()
            **kwargs: arguments to pass to _postprocessing()

        Returns:
            object: as returned by _postprocessing()
        """
        result = super(LightBlueGenericSelection, self).update()
        return self._postprocessing(result, *args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Postprocessing wrapper over delete method.

        Args:
            *args: arguments to pass to _postprocessing()
            **kwargs: arguments to pass to _postprocessing()

        Returns:
            object: as returned by _postprocessing()
        """
        result = super(LightBlueGenericSelection, self).delete()
        return self._postprocessing(result, *args, **kwargs)

    # .insert() method is public, without any changes

    @staticmethod
    def get_selector_query(data, primary_keys):
        """
        Get a query to identify an entity.

        This method receives an item and PKs and returns a query.

        Q: Why this method is static if we have PKs in constructor?
        A: This method have to be accessible from the classmethod
        insert_and_select(), which has no access to the attribute.

        Q: Why PKs are not class attributes?
        A: Even if we will define PKs as a class attribute and
        override them in child class - overridden value won't be
        accessible from the classmethod.

        Args:
            data (dict): data as stored in LightBlue
            primary_keys (dict):
                - key is a filed name
                - value is dpath selector

        Returns:
            list: query (as pairs) to identify a specific entity.
        """
        query = []
        for key, selector in primary_keys.items():
            query.append((key, dpath.util.get(data, selector), ))
        return query

    # TODO: design a classmethod to insert item and return selector

    # .add_raw_query() method is public from LightBlueQuery
    def with_lb_id(self):
        """
        Include LB _id field

        Allows method chaining, returns self.
        """
        self._add_to_projection('_id')
        return self

    def filter_created_by(self, service):
        """
        Select items created by specific service.

        Args:
            service (str): service name

        Allows method chaining, returns self.
        """
        self._add_to_query(createdBy=service)
        return self

    @property
    def first(self):
        """
        Get the first element of response for a query.

        Returns:
            (dict, None):
                first item if available, None otherwise
        """
        return self.find(
            count=(1, ),
            selector='/processed/0',
            fallback=None)

    @property
    def exist(self):
        """
        Check if items are available for the query.

        Returns:
            bool: True if items exist, False otherwise
        """
        result = self.find()
        if result is None:
            return False
        else:
            return result['matchCount'] > 0

    @property
    def all(self):
        """
        Get all elements of response for given quert

        Returns:
            (list / []) - list of found items, empty list otherwise

        """
        return self.find(
            count=(1, ),
            selector='/processed',
            fallback=[]
        )

    def update_with(self, data):
        """
        Update item with given data

        Args:
            data (dict): update data

        Returns:
            dict: response as specified in the projection

        """
        self._add_to_update(_set=data)
        return self.update()

    def unset_fields(self, fields):
        """
        Unset fields.

        Skips update if fields is empty.

        Args:
            fields (list): list of sorted fields, for example:
                ['field.3', 'field.1']

        Returns:
            dict: response from lightblue update query
            int: count of removed fields, 0 if update failed

        """
        if not fields:
            return None, 0
        # check response is included
        self._add_to_update(unset=fields)
        response = self.update()
        return response, 0 if response is None else len(fields)



"""
Custom class of BeanBag
"""

import logging

from beanbag.v2 import BeanBag, BeanBagException

LOGGER = logging.getLogger('lightblue')


class CustomBeanBag(~BeanBag):
    """
    Custom BeanBag class which prints error content
    """
    def __init__(self, *args, **kwargs):
        super(~CustomBeanBag, self).__init__(*args, **kwargs)

    def decode(self, response):
        """
        Custom decode function which prints response when request gets error
        Args:
            response (object): response from server

        Returns: response AttrDict

        """
        if response.status_code < 200 or response.status_code >= 300:
            LOGGER.warning("Invalid response code (%s) "
                           "received from service: %s",
                           response.status_code, self.base_url)
            # print response content
            raise BeanBagException(
                response, "Bad response code: %d\nResponse: %s" % (
                    response.status_code, response.content))

        return super(~CustomBeanBag, self).decode(response)



import requests

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


def retry_session(session=None):
    """
    Retry session in case it failed
    More info: https://github.com/mikem23/keepalive-race

    Args:
        session (object): already created session - in case it is missing new
        session is created
    Returns:
        session object with retry settings
    """
    if not session:
        session = requests.Session()
    retry = Retry(
        total=5,
        read=5,
        connect=5,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504),
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



import os
import io

from setuptools import setup

PROJECT_ROOT, _ = os.path.split(__file__)

NAME = 'python-lightblue'
EMAILS = 'araszka@redhat.com, mbirger@redhat.com'
AUTHORS = 'Ales Raszka, Mark Birger'
VERSION = '1.0.1'

URL = 'https://github.com/redhat-aqe/python-lightblue'
LICENSE = 'GPLv3'


SHORT_DESCRIPTION = 'A Python API for Lightblue database.'

with io.open(os.path.join(PROJECT_ROOT, 'README.md'), encoding="utf-8") as f:
    DESCRIPTION = f.read()

INSTALL_REQUIRES = open(os.path.join(PROJECT_ROOT, 'requirements.txt')). \
        read().splitlines()
TEST_REQUIRES = open(os.path.join(PROJECT_ROOT, 'build_requirements.txt')). \
        read().splitlines()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHORS,
    author_email=EMAILS,
    package_dir={'': 'src'},
    packages=[
        'lightblue',
        ],
    install_requires=INSTALL_REQUIRES,
    test_suite='nose.collector',
    tests_require=TEST_REQUIRES,
    url=URL,
    description=SHORT_DESCRIPTION,
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    license=LICENSE,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from unittest import TestCase
import requests

from lightblue.service import LightBlueService

try:
    from unittest.mock import Mock, patch
except ImportError:
    from mock import Mock, patch


class TestLightBlueService(TestCase):
    """
    Test cases for LightBlueService class
    """

    test_docstring_prefix = "Service - "

    def shortDescription(self):  # noqa
        """Override nosetest docstrings."""
        doc = self.test_docstring_prefix + self._testMethodDoc
        return doc or None

    def setUp(self):
        self.data_url = 'http:/fake.lb.com/rest/data'
        self.metadata_url = 'http:/fake.lb.com/rest/metadata'

        self.service = LightBlueService(
            self.data_url, self.metadata_url, ssl_certificate='path/to/cert'
        )

    def test_custom_session(self):
        session = requests.session()
        service = LightBlueService(
            self.data_url,
            self.metadata_url,
            ssl_certificate='path/to/cert',
            custom_session=session,
        )
        self.assertEqual(service.session, session)

    def test_log_response_json(self):
        """
        Test response message is complete
        """
        response = Mock()
        response.json.return_value = {
            'status': 'FAILED',
            'matchCount': 0,
            'modifiedCount': 0,
            'dataErrors': None,
            'data': ['object'],
        }
        response.status_code = 200
        result = self.service.log_response(response)
        expected_result = {
            'status': 'FAILED',
            'matchCount': 0,
            'modifiedCount': 0,
            'dataErrors': None,
            'errors': None,
            'statusCode': 200,
        }
        self.assertEqual(result, expected_result)

    def test_log_response_no_json(self):
        """
        Test response message is complete - failed
        """
        response = Mock()
        response.json.side_effect = ValueError('err')
        response.text = 'response'
        response.status_code = 500
        result = self.service.log_response(response)
        expected_result = {'text_response': 'response', 'statusCode': 500}
        self.assertEqual(result, expected_result)

    @patch('requests.Session.get')
    def test_get_schema(self, mock_get):
        """
        Test of getting schema
        """
        schema = {'key': 'value'}
        mock_get.return_value.json.return_value = schema
        result = self.service.get_schema('entity', 'version')
        self.assertEqual(result, schema)

    @patch('requests.Session.put')
    def test_insert_data(self, mock_put):
        """
        Test of inserting data
        """
        data = 'object'
        resp_data = {
            'status': 'COMPLETE',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_put.return_value.json.return_value = resp_data
        mock_put.return_value.status_code = 200
        result = self.service.insert_data('entity', 'version', data)
        call_args = mock_put.call_args
        self.assertEqual(
            call_args[0][0], '{}/insert/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertEqual(result, resp_data)

    @patch('requests.Session.put')
    def test_insert_data_failed(self, mock_put):
        """
        Test of inserting data - failed request
        """
        data = 'object'
        resp_data = {
            'status': 'FAILED',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_put.return_value.json.return_value = resp_data
        mock_put.return_value.status_code = 500
        result = self.service.insert_data('entity', 'version', data)
        call_args = mock_put.call_args
        self.assertEqual(
            call_args[0][0], '{}/insert/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertIsNone(result)

    @patch('requests.Session.post')
    def test_delete_data(self, mock_post):
        """
        Test of deleting data
        """
        data = 'object'
        resp_data = {
            'status': 'COMPLETE',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 200
        result = self.service.delete_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/delete/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertEqual(result, resp_data)

    @patch('requests.Session.post')
    def test_delete_data_failed(self, mock_post):
        """
        Test of deleting data - failed request
        """
        data = 'object'
        resp_data = {
            'status': 'FAILED',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 500
        result = self.service.delete_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/delete/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertIsNone(result)

    @patch('requests.Session.post')
    def test_update_data(self, mock_post):
        """
        Test of updating data
        """
        data = 'object'
        resp_data = {
            'status': 'COMPLETE',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 200
        result = self.service.update_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/update/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertEqual(result, resp_data)

    @patch('requests.Session.post')
    def test_update_data_failed(self, mock_post):
        """
        Test of updating data - failed request
        """
        data = 'object'
        resp_data = {
            'status': 'FAILED',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 500
        result = self.service.update_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/update/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertIsNone(result)

    @patch('requests.Session.post')
    def test_find_data(self, mock_post):
        """
        Test of finding data
        """
        data = 'object'
        resp_data = {
            'status': 'COMPLETE',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 200
        result = self.service.find_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/find/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertEqual(result, resp_data)

    @patch('requests.Session.post')
    def test_find_data_failed(self, mock_post):
        """
        Test of finding data - failed request
        """
        data = 'object'
        resp_data = {
            'status': 'FAILED',
            'matchCount': 1,
            'modifiedCount': 1,
            'data': [data],
        }
        mock_post.return_value.json.return_value = resp_data
        mock_post.return_value.status_code = 500
        result = self.service.find_data('entity', 'version', data)
        call_args = mock_post.call_args
        self.assertEqual(
            call_args[0][0], '{}/find/entity/version'.format(self.data_url)
        )
        self.assertEqual(call_args[1], {'json': data})
        self.assertIsNone(result)



"""LightBlueQuery implementation."""


class IncompleteQuery(Exception):
    """Attempt to execute a query to LightBlue (w/o proper data)."""

    pass


class LockedQuery(Exception):
    """Attempt to execute non-find query to LightBlue second+ time."""

    pass


class LightBlueQuery(object):
    """
    LightBlue query representation.

    (in both executed and non-executed state)

    Attributes:
        interface (lightblue.entity.LightBlueEntity):
            wrapper to query a LightBlue method
    """

    def __init__(self,
                 interface,
                 *args, **kwargs):
        """
        Initialize a LightBlueQuery object.

        Args:
            interface (lightblue.entity.LightBlueEntity):
                reference to LightBlueEntity object,
                created in collection class
                (to avoid session creation on each initialization)
            *args: list of pairs/triples with initial query
            **kwargs: dict with initial query (equals)
        """
        super(LightBlueQuery, self).__init__()
        # lock to avoid calling multiple times
        self._locked = False
        # setup interface to LB
        self.interface = interface
        # query is a list of triples (field, op, rvalue)
        queries_pairs = self._kwargs_to_pairs(**kwargs)
        args = list(map(self._query_pairs_to_triples, args))
        queries_pairs = list(map(self._query_pairs_to_triples, queries_pairs))
        self._queries = args + queries_pairs
        self._raw_queries = []
        # projections is a list of fields
        self._projections = []
        # projections_recursive is a list of fields with recursive=True
        self._projections_recursive = []
        # update set is a key-value dict
        self._update_set = {}
        # update unset is a fields list
        self._update_unset = []
        # update append is a dict
        self._update_append = {}

    @property
    def kwargs_aliases(self):
        """
        Return aliases for LightBlue fields.

        It is better to override them in child class, than pass
        another keyword argument and mix it with kwargs.

        Returns:
            dict: empty dict, no aliases by default
        """
        return {}

    # helper methods to parse args
    @staticmethod
    def _query_pairs_to_triples(query):
        """
        Transform pairs (no changes in case of triples).

        Skips triples by default.

        Args:
            query (tuple): pair (field, rvalue)
                or triple (field, relation, rvalue)

        Returns:
            tuple: triple (field, relation, rvalue)
        """
        if len(query) == 3:
            return query
        else:
            return query[0], '=', query[1]

    def _kwargs_to_pairs(self, **kwargs):
        """
        Transform kwargs to pairs.

        Args:
            **kwargs: query values

        Returns:
            tuple: (field, rvalue)
        """
        pairs = []
        for key, value in kwargs.items():
            if key in self.kwargs_aliases:
                pairs.append((self.kwargs_aliases[key], value, ))
            else:
                pairs.append((key, value, ))
        return pairs

    # internal methods to update query/projection/update
    def _add_to_query(self, *args, **kwargs):
        """
        Transform and save query into the attribute.

        Args:
            *args: list of pairs/triples with query
            **kwargs: dict with query (equals)
        """
        queries_pairs = self._kwargs_to_pairs(**kwargs)
        args = list(map(self._query_pairs_to_triples, args))
        queries_pairs = list(map(self._query_pairs_to_triples, queries_pairs))
        self._queries += args + queries_pairs

    def add_raw_query(self, query):
        """
        Save raw query to the attribute.

        This is a public method, because query can be constructed
        on the abstraction level of a collection.

        Args:
            query (dict): raw LightBlue query
        """
        self._raw_queries.append(query)

    def _add_to_projection(self, *args, **kwargs):
        """
        Transform and save projection into the attribute.

        Args:
            *args: field names
            **kwargs: dict with recursive key (fields with recursive=True)
        """
        recursive = kwargs.pop('recursive', [])
        self._projections += list(args)
        self._projections_recursive += recursive

    def _add_to_update(self,
                       _set=None,
                       unset=None,
                       append=None):
        """
        Save update data to the attribute.

        Args:
            _set (dict, optional):  part of the update query
            unset (list, optional):  part of the update query
            append (dict, optional):  part of the update query
        """
        if _set is not None:
            self._update_set.update(_set)
        if unset is not None:
            self._update_unset += unset
        if append is not None:
            self._update_append.update(append)

    def _refresh(self):
        """
        Clear projection and update.

        Leaves query as is.
        """
        self._projections = []
        self._projections_recursive = []
        self._update_set = {}
        self._update_unset = []
        self._update_append = {}
        # unlock future queries
        self._locked = False

    @staticmethod
    def _and(*args):
        """
        Logical AND for multiple queries.

        Args:
            *args: list of queries

        Returns:
            dict: LightBlue query
        """
        return {
            "$and": list(args)
        }

    # properties, that construct a request
    @property
    def _has_query(self):
        """
        Object has at least one query specified.

        Returns:
            bool: True if has, False otherwise
        """
        return len(self._queries) > 0 or \
            len(self._raw_queries) > 0

    @property
    def _query(self):
        """
        Construct LightBlue query.

        Returns:
            dict: LightBlue query
        """
        dict_queries = [{
            'field': query[0],
            'op': query[1],
            'rvalue': query[2]
        } for query in self._queries]
        return self._and(*(dict_queries + self._raw_queries))

    @property
    def _has_projection(self):
        """
        Object has at least one projection specified.

        Returns:
            bool: True if has, False otherwise
        """
        return len(self._projections) > 0 or \
            len(self._projections_recursive) > 0

    @property
    def _projection(self):
        """
        Construct LightBlue projection.

        Returns:
            dict: LightBlue projection
        """
        normal = [{
            'field': field,
            'include': True
        } for field in self._projections]
        recursive = [{
            'field': field,
            'include': True,
            'recursive': True
        } for field in self._projections_recursive]
        return normal + recursive

    @property
    def _has_update(self):
        """
        Object has at least one update specified.

        Returns:
            bool: True if has, False otherwise
        """
        return bool(self._update_set) or \
            len(self._update_unset) > 0 or \
            bool(self._update_append)

    @property
    def _update(self):
        """
        Construct LightBlue update.

        Returns:
            dict: LightBlue update
        """
        result = {}
        if self._update_set:
            result['$set'] = self._update_set
        if self._update_unset:
            result['$unset'] = self._update_unset
        if self._update_append:
            result['$append'] = self._update_append
        return result

    def find(self):
        """
        Execute find call to LightBlue.

        Returns: raw response from LB
        """
        if self._has_query:
            if self._has_projection:
                return self.interface.find_item(
                    self._query,
                    projection=self._projection)
            else:
                return self.interface.find_item(self._query)
        else:
            if self._has_projection:
                return self.interface.find_all(
                    projection=self._projection)
            else:
                return self.interface.find_all()

    def update(self):
        """
        Execute update call to LightBlue.

        Returns: raw response from LB

        Raises:
            IncompleteQuery: in case of unspecified update
            LockedQuery: in case of second+ call
        """
        if self._locked:
            raise LockedQuery()
        if self._has_query and self._has_update:
            self._locked = True
            return self.interface.update_item(
                self._query, self._update)
        else:
            raise IncompleteQuery()

    def delete(self):
        """
        Execute delete call to LightBlue.

        Returns:
            object: raw response from LB

        Raises:
            IncompleteQuery: in case of unspecified query
                (can't delete all items by default)
            LockedQuery: in case of second+ call
        """
        if self._locked:
            raise LockedQuery()
        if self._has_query:
            self._locked = True
            return self.interface.delete_item(self._query)
        else:
            raise IncompleteQuery()

    @staticmethod
    def insert(data, interface):
        """
        Insert a new item to the collection.

        Args:
            data (dict): data as required by a schema
            interface (metaxor.lightblue.base.LightBlueEntity):
                wrapper to query LightBlue methods

        Returns:
            dict: raw response from LB
        """
        return interface.insert_data(data)
